﻿/*
 * UPF Parser and XML Exporter
 * 
 * written by Christian B. Mendl 2012
 * 
 * Reference: Quantum Espresso Unified Pseudopotential Format (UPF)
 * http://www.quantum-espresso.org/?page_id=269
 */

package ParseUPF;

import java.io.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import org.w3c.dom.*;


class ExportXML
{
	private void ExportHeader(Document doc, UPFData.PPHeader header, Element node)
	{
		// version number
		{
			Element n = doc.createElement("n");
			n.setTextContent(Integer.toString(header.n));
			node.appendChild(n);
		}

		{
			Element psd = doc.createElement("psd");
			psd.setTextContent(header.psd);
			node.appendChild(psd);
		}

		{
			Element type = doc.createElement("type");
			type.setTextContent(header.type.toString());
			node.appendChild(type);
		}

		{
			Element nlcc = doc.createElement("nlcc");
			nlcc.setTextContent(Boolean.toString(header.nlcc));
			node.appendChild(nlcc);
		}

		{
			Element dft = doc.createElement("dft");

			Element exch = doc.createElement("exch");
			exch.setTextContent(header.dft.exch);
			dft.appendChild(exch);

			Element correlations = doc.createElement("correlations");
			correlations.setTextContent(header.dft.correlations);
			dft.appendChild(correlations);

			Element exchGradCorrection = doc.createElement("exchGradCorrection");
			exchGradCorrection.setTextContent(header.dft.exchGradCorrection);
			dft.appendChild(exchGradCorrection);

			Element correlationsGradCorrection = doc.createElement("correlationsGradCorrection");
			correlationsGradCorrection.setTextContent(header.dft.correlationsGradCorrection);
			dft.appendChild(correlationsGradCorrection);

			node.appendChild(dft);
		}

		{
			Element zp = doc.createElement("zp");
			zp.setTextContent(Double.toString(header.zp));
			node.appendChild(zp);
		}

		{
			Element etotps = doc.createElement("etotps");
			etotps.setTextContent(Double.toString(header.etotps));
			node.appendChild(etotps);
		}

		{
			Element ecutwfc = doc.createElement("ecutwfc");
			ecutwfc.setTextContent(Double.toString(header.ecutwfc));
			node.appendChild(ecutwfc);

			Element ecutrho = doc.createElement("ecutrho");
			ecutrho.setTextContent(Double.toString(header.ecutrho));
			node.appendChild(ecutrho);
		}

		{
			Element lmax = doc.createElement("lmax");
			lmax.setTextContent(Integer.toString(header.lmax));
			node.appendChild(lmax);
		}

		{
			Element mesh = doc.createElement("mesh");
			mesh.setTextContent(Integer.toString(header.mesh));
			node.appendChild(mesh);
		}

		{
			Element natwfc = doc.createElement("natwfc");
			natwfc.setTextContent(Integer.toString(header.natwfc));
			node.appendChild(natwfc);

			Element nbeta = doc.createElement("nbeta");
			nbeta.setTextContent(Integer.toString(header.nbeta));
			node.appendChild(nbeta);
		}

		{
			Element wavefunctions = doc.createElement("wavefunctions");

			for (UPFData.PPHeader.Wavefunction w: header.wavefunctions)
			{
				Element wNode = doc.createElement("wavefunction");

				Attr els = doc.createAttribute("els");
				els.setValue(w.els.toString().substring(1));
				wNode.setAttributeNode(els);

				Attr lchi = doc.createAttribute("lchi");
				lchi.setValue(Integer.toString(w.lchi));
				wNode.setAttributeNode(lchi);

				Attr oc = doc.createAttribute("oc");
				oc.setValue(Double.toString(w.oc));
				wNode.setAttributeNode(oc);

				wavefunctions.appendChild(wNode);
			}

			node.appendChild(wavefunctions);
		}
	}

	private void ExportMesh(Document doc, UPFData.PPMesh mesh, Element node)
	{
		{
			Element radial = doc.createElement("PP_R");

			String innerText = "";
			for (double r: mesh.r)
			{
				innerText += r + "\n";
			}
			radial.setTextContent(innerText);

			node.appendChild(radial);
		}

		{
			Element factors = doc.createElement("PP_RAB");
			
			String innerText = "";
			for (double r: mesh.rab)
			{
				innerText += r + "\n";
			}
			factors.setTextContent(innerText);

			node.appendChild(factors);
		}
	}

	private void ExportLocal(Document doc, UPFData.PPLocal local, Element node)
	{
		String innerText = "";
		for (double v: local.vloc)
		{
			innerText += v + "\n";
		}
		node.setTextContent(innerText);
	}

	private void ExportNLCC(Document doc, UPFData.PPNLCC nlcc, Element node)
	{
		String innerText = "";
		for (double rho: nlcc.rho_atc)
		{
			innerText += rho + "\n";
		}
		node.setTextContent(innerText);
	}

	private void ExportNonLocal(Document doc, UPFData.PPNonLocal nonLocal, Element node)
	{
		for (UPFData.PPNonLocal.PPBeta betaElem: nonLocal.betaList)
		{
			Element nodeBeta = doc.createElement("PP_BETA");

			{
				Attr lll = doc.createAttribute("lll");
				lll.setValue(Integer.toString(betaElem.lll));
				nodeBeta.setAttributeNode(lll);
			}

			{
				Attr kkbeta = doc.createAttribute("kkbeta");
				kkbeta.setValue(Integer.toString(betaElem.kkbeta));
				nodeBeta.setAttributeNode(kkbeta);
			}

			{
				String innerText = "";
				for (double r: betaElem.beta)
				{
					innerText += r + "\n";
				}
				nodeBeta.setTextContent(innerText);
			}

			node.appendChild(nodeBeta);
		}

		Element nodeD = doc.createElement("PP_DIJ");
		String innerText = "";
		for (int i = 0; i < nonLocal.D.length; i++)
		{
			for (int j = 0; j < nonLocal.D[0].length; j++)
			{
				innerText += nonLocal.D[i][j] + " ";
			}
			innerText += "; ";
		}
		innerText = innerText.substring(0, innerText.length() - 2);		// remove trailing '; '
		nodeD.setTextContent(innerText);

		node.appendChild(nodeD);
	}

	private void ExportPSWFC(Document doc, UPFData.PPPSWFC pswfc, Element node)
	{
		for (UPFData.PPPSWFC.Wavefunction w: pswfc.wavefunctions)
		{
			Element wavefunction = doc.createElement("wavefunction");

			{
				Attr els = doc.createAttribute("els");
				els.setValue(w.els.toString().substring(1));
				wavefunction.setAttributeNode(els);
			}

			{
				Attr lchi = doc.createAttribute("lchi");
				lchi.setValue(Integer.toString(w.lchi));
				wavefunction.setAttributeNode(lchi);
			}

			{
				Attr oc = doc.createAttribute("oc");
				oc.setValue(Double.toString(w.oc));
				wavefunction.setAttributeNode(oc);
			}

			{
				String innerText = "";
				for (double z: w.chi)
				{
					innerText += z + "\n";
				}

				Element chi = doc.createElement("chi");
				chi.setTextContent(innerText);
				wavefunction.appendChild(chi);
			}

			node.appendChild(wavefunction);
		}			
	}

	private void ExportRhoAtom(Document doc, UPFData.PPRhoAtom rho, Element node)
	{
		String innerText = "";
		for (double r: rho.rho_at)
		{
			innerText += r + "\n";
		}
		
		node.setTextContent(innerText);
	}


	public void Export(UPFData data, String filename)
	{
		try
		{
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
 
			Document doc = docBuilder.newDocument();
	
			// root element
			Element root = doc.createElement("UPF");
			doc.appendChild(root);
	
	
			//Document doc = new Document();
	
			//XmlElement root = doc.createElement("UPF");
			//doc.appendChild(root);
	
			{
				Element nodeHeader = doc.createElement("PP_HEADER");
				ExportHeader(doc, data.header, nodeHeader);
				root.appendChild(nodeHeader);
			}
	
			{
				Element nodeMesh = doc.createElement("PP_MESH");
				ExportMesh(doc, data.mesh, nodeMesh);
				root.appendChild(nodeMesh);
			}
	
			{
				Element nodeNLCC = doc.createElement("PP_NLCC");
				ExportNLCC(doc, data.nlcc, nodeNLCC);
				root.appendChild(nodeNLCC);
			}
	
			{
				Element nodeLocal = doc.createElement("PP_LOCAL");
				ExportLocal(doc, data.local, nodeLocal);
				root.appendChild(nodeLocal);
			}
	
			{
				Element nodeNonLocal = doc.createElement("PP_NONLOCAL");
				ExportNonLocal(doc, data.nonLocal, nodeNonLocal);
				root.appendChild(nodeNonLocal);
			}
	
			{
				Element nodePSWFC = doc.createElement("PP_PSWFC");
				ExportPSWFC(doc, data.pswfc, nodePSWFC);
				root.appendChild(nodePSWFC);
			}
	
			{
				Element nodeRhoAtom = doc.createElement("PP_RHOATOM");
				ExportRhoAtom(doc, data.rhoAtom, nodeRhoAtom);
				root.appendChild(nodeRhoAtom);
			}
	

			// save to XML file
			Transformer transformer = TransformerFactory.newInstance().newTransformer();
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(new File(filename));
			transformer.transform(source, result);
		}
		catch (ParserConfigurationException e)
		{
			e.printStackTrace();

		}
		catch (TransformerException e)
		{
			e.printStackTrace();
		}
	}
}
