﻿/*
 * UPF Parser and XML Exporter
 * 
 * written by Christian B. Mendl 2012
 * 
 * Reference: Quantum Espresso Unified Pseudopotential Format (UPF)
 * http://www.quantum-espresso.org/?page_id=269
 */

package ParseUPF;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;


class Parser
{
	//private static readonly char[] separator = new char[] { ' ', '\t' };
	private final static String separator = "\\s+"; 

	private static boolean CompareStringArrays(String[] a, String[] b)
	{
		if (a == null || b == null)
		{
			return false;
		}
		if (a.length != b.length)
		{
			return false;
		}

		for (int i = 0; i < a.length; i++)
		{
			if (!a[i].equalsIgnoreCase(b[i]))
			{
				return false;
			}
		}

		return true;
	}


	private UPFData.PPHeader ParseHeader(BufferedReader reader) throws IOException
	{
		UPFData.PPHeader hdr = new UPFData.PPHeader();

		// version number
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"version", "number"}))
			{
				System.err.println("Expecting 'Version Number'");
			}
			hdr.n = Integer.parseInt(items[0]);
		}

		// element
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"Element"}))
			{
				System.err.println("Expecting 'Element'");
			}
			hdr.psd = items[0];
		}

		// type
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (items.length < 2)
			{
				System.err.println("Could not read pseudopotential type");
			}
			hdr.type = UPFData.PPHeader.Type.valueOf(items[0].toUpperCase());
		}

		// non-linear core correction?
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"nonlinear", "core", "correction"}))
			{
				System.err.println("Could not determine non-linear core correction");
			}
			if (items[0].equalsIgnoreCase("T") || items[0].equalsIgnoreCase("true"))
			{
				hdr.nlcc = true;
			}
			else if (items[0].equalsIgnoreCase("F") || items[0].equalsIgnoreCase("false"))
			{
				hdr.nlcc = false;
			}
			else
			{
				System.err.println("Unknown non-linear core correction");
			}
		}

		// exchange-correlation functional
		{
			hdr.dft = new UPFData.PPHeader.ExchangeCorrelationFunctional();

			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (items.length < 5)
			{
				System.err.println("Could not determine exchange-correlation");
			}
			hdr.dft.exch                       = items[0];
			hdr.dft.correlations               = items[1];
			hdr.dft.exchGradCorrection         = items[2];
			hdr.dft.correlationsGradCorrection = items[3];
		}

		// Z valence
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"Z", "valence"}))
			{
				System.err.println("Could not determine Z valence");
			}
			hdr.zp = Double.parseDouble(items[0]);
		}

		// total energy
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"total", "energy"}))
			{
				System.err.println("Could not determine total energy");
			}
			hdr.etotps = Double.parseDouble(items[0]);
		}

		// suggested cut-off for wfc and rho
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 2, items.length), new String[] {"suggested", "cutoff", "for", "wfc", "and", "rho"}))
			{
				System.err.println("Could not determine suggested cutoff for wfc and rho");
			}
			hdr.ecutwfc = Double.parseDouble(items[0]);
			hdr.ecutrho = Double.parseDouble(items[1]);
		}

		// max angular momentum component
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"max", "angular", "momentum", "component"}))
			{
				System.err.println("Could not determine max angular momentum component");
			}
			hdr.lmax = Integer.parseInt(items[0]);
		}

		// number of points in mesh
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 1, items.length), new String[] {"number", "of", "points", "in", "mesh"}))
			{
				System.err.println("Could not determine number of points in mesh");
			}
			hdr.mesh = Integer.parseInt(items[0]);
		}

		// number of wavefunctions, projectors
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(Arrays.copyOfRange(items, 2, items.length), new String[] {"Number", "of", "Wavefunctions,", "Number", "of", "Projectors"}))
			{
				System.err.println("Could not determine number of wavefunctions and projectors");
			}
			hdr.natwfc = Integer.parseInt(items[0]);
			hdr.nbeta  = Integer.parseInt(items[1]);
		}

		// wavefunctions
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (!CompareStringArrays(items, new String[] {"wavefunctions", "nl", "l", "occ"}))
			{
				System.err.println("Error while reading wavefunctions");
			}

			hdr.wavefunctions = new ArrayList<UPFData.PPHeader.Wavefunction>(hdr.natwfc);
			for (int i = 0; i < hdr.natwfc; i++)
			{
				line = reader.readLine();
				if (line == null)
				{
					System.err.println("Could not read line");
				}
				line = line.trim();
				items = line.split(separator);
				if (items.length != 3)
				{
					System.err.println("Could not read wavefunction");
				}
				UPFData.PPHeader.Wavefunction w = new UPFData.PPHeader.Wavefunction();
				w.els = UPFData.OrbitalLabel.valueOf(("_" + items[0]).toUpperCase());
				w.lchi = Integer.parseInt(items[1]);
				w.oc   = Double.parseDouble(items[2]);

				hdr.wavefunctions.add(w);
			}
		}

		// final closing tag
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}
			line = line.trim();
			if (!line.equalsIgnoreCase("</PP_HEADER>"))
			{
				System.err.println("Expecting </PP_HEADER>");
			}
		}

		return hdr;
	}


	private UPFData.PPMesh ParseMesh(BufferedReader reader) throws IOException
	{
		UPFData.PPMesh m = new UPFData.PPMesh();
		m.r = new ArrayList<Double>();
		m.rab = new ArrayList<Double>();

		while (true)
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}

			line = line.trim();
			if (line.equalsIgnoreCase("<PP_R>"))
			{
				while (true)
				{
					line = reader.readLine();
					if (line == null)
					{
						System.err.println("Could not read line inside '<PP_R>'");
					}

					line = line.trim();
					if (line.equalsIgnoreCase("</PP_R>"))
					{
						// radial mesh points done
						break;
					}

					String[] items = line.split(separator);
					for (String item: items)
						m.r.add(Double.parseDouble(item));
				}
			}
			else if (line.equalsIgnoreCase("<PP_RAB>"))
			{
				while (true)
				{
					line = reader.readLine();
					if (line == null)
					{
						System.err.println("Could not read line inside '<PP_RAB>'");
					}

					line = line.trim();
					if (line.equalsIgnoreCase("</PP_RAB>"))
					{
						// mesh integration factors done
						break;
					}

					String[] items = line.split(separator);
					for (String item: items)
						m.rab.add(Double.parseDouble(item));
				}
			}
			else if (line.equalsIgnoreCase("</PP_MESH>"))
			{
				// finished
				break;
			}
			else
			{
				System.err.println("Expecting '<PP_R>' or '<PP_RAB>'");
			}
		}

		if (m.r.size() != m.rab.size())
		{
			System.err.println("Number of radial mesh points and integration factors must agree");
		}

		return m;
	}

	private UPFData.PPNLCC ParseNLCC(BufferedReader reader) throws IOException
	{
		UPFData.PPNLCC nlcc = new UPFData.PPNLCC();
		nlcc.rho_atc = new ArrayList<Double>();

		while (true)
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line inside non-linear core correction");
			}

			line = line.trim();
			if (line.equalsIgnoreCase("</PP_NLCC>"))
			{
				break;
			}

			String[] items = line.split(separator);
			for (String item: items)
				nlcc.rho_atc.add(Double.parseDouble(item));
		}

		return nlcc;
	}

	private UPFData.PPLocal ParseLocal(BufferedReader reader) throws IOException
	{
		UPFData.PPLocal loc = new UPFData.PPLocal();
		loc.vloc = new ArrayList<Double>();

		while (true)
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line inside local potential");
			}

			line = line.trim();
			if (line.equalsIgnoreCase("</PP_LOCAL>"))
			{
				break;
			}

			String[] items = line.split(separator);
			for (String item: items)
				loc.vloc.add(Double.parseDouble(item));
		}

		return loc;
	}

	private UPFData.PPNonLocal ParseNonLocal(BufferedReader reader) throws IOException
	{
		UPFData.PPNonLocal nloc = new UPFData.PPNonLocal();
		nloc.betaList = new ArrayList<UPFData.PPNonLocal.PPBeta>();

		while (true)
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line");
			}

			line = line.trim();
			if (line.equalsIgnoreCase("<PP_BETA>"))
			{
				UPFData.PPNonLocal.PPBeta beta = new UPFData.PPNonLocal.PPBeta();

				line = reader.readLine();
				if (line == null)
				{
					System.err.println("Could not read line inside '<PP_BETA>'");
				}
				line = line.trim();
				String[] items = line.split(separator);
				if (items.length != 4 || !CompareStringArrays(Arrays.copyOfRange(items, 2, items.length), new String[] { "Beta", "L" }))
				{
					System.err.println("Error while reading beta element");
				}
				if (Integer.parseInt(items[0]) != nloc.betaList.size() + 1)
				{
					System.err.println("Wrong beta element index");
				}
				beta.lll = Integer.parseInt(items[1]);

				line = reader.readLine();
				if (line == null)
				{
					System.err.println("Could not read line inside '<PP_BETA>'");
				}
				line = line.trim();
				beta.kkbeta = Integer.parseInt(line);

				beta.beta = new ArrayList<Double>();
				while (true)
				{
					line = reader.readLine();
					if (line == null)
					{
						System.err.println("Could not read line inside '<PP_BETA>'");
					}

					line = line.trim();
					if (line.equalsIgnoreCase("</PP_BETA>"))
					{
						break;
					}

					items = line.split(separator);
					for (String item: items)
						beta.beta.add(Double.parseDouble(item));
				}

				if (beta.beta.size() != beta.kkbeta)
				{
					System.err.println("Number of beta radial points does not match");
				}

				nloc.betaList.add(beta);
			}
			else if (line.equalsIgnoreCase("<PP_DIJ>"))
			{
				if (nloc.betaList.isEmpty())
				{
					System.err.println("Unknown number of beta in 'PP_DIJ'");
				}

				nloc.D = new double[nloc.betaList.size()][nloc.betaList.size()];

				// number of nonzero Dij
				line = reader.readLine();
				if (line == null)
				{
					System.err.println("Could not read line inside '<PP_DIJ>'");
				}
				line = line.trim();
				String[] items = line.split(separator);
				if (items.length != 5)
				{
					System.err.println("Error while reading PP_DIJ");
				}
				int nd = Integer.parseInt(items[0]);

				for (int i = 0; i < nd; i++)
				{
					line = reader.readLine();
					if (line == null)
					{
						System.err.println("Could not read line inside '<PP_DIJ>'");
					}
					line = line.trim();
					items = line.split(separator);
					if (items.length != 3)
					{
						System.err.println("Error while reading PP_DIJ");
					}

					nloc.D[Integer.parseInt(items[0]) - 1][Integer.parseInt(items[1]) - 1] = Double.parseDouble(items[2]);
				}

				line = reader.readLine();
				if (line == null)
				{
					System.err.println("Could not read line inside '<PP_DIJ>'");
				}
				line = line.trim();
				if (!line.equalsIgnoreCase("</PP_DIJ>"))
				{
					System.err.println("Expecting '</PP_DIJ>'");
				}
			}
			else if (line.equalsIgnoreCase("<PP_QIJ>"))
			{
				System.err.println("'</PP_QIJ>' not implemented yet");
			}
			else if (line.equalsIgnoreCase("</PP_NONLOCAL>"))
			{
				// finished
				break;
			}
			else
			{
				System.err.println("Expecting '<PP_BETA>', '<PP_DIJ>' or '<PP_QIJ>'");
			}
		}

		return nloc;
	}

	private UPFData.PPPSWFC ParsePSWFC(BufferedReader reader, int natwfc, int mesh) throws IOException
	{
		UPFData.PPPSWFC pswfc = new UPFData.PPPSWFC();
		pswfc.wavefunctions = new ArrayList<UPFData.PPPSWFC.Wavefunction>();

		String line;

		for (int i = 0; i < natwfc; i++)
		{
			UPFData.PPPSWFC.Wavefunction w = new UPFData.PPPSWFC.Wavefunction();

			line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line inside PSWFC");
			}
			line = line.trim();
			String[] items = line.split(separator);
			if (items.length != 4 || !items[3].equalsIgnoreCase("Wavefunction"))
			{
				System.err.println("Error while reading wavefunction in PSWFC");
			}
			w.els = UPFData.OrbitalLabel.valueOf(("_" + items[0]).toUpperCase());
			w.lchi = Integer.parseInt(items[1]);
			w.oc   = Double.parseDouble(items[2]);

			// read i-th radial atomic (pseudo-)orbital (radial part of the KS equation, multiplied by r)
			w.chi = new ArrayList<Double>(mesh);
			while (w.chi.size() < mesh)
			{
				line = reader.readLine();
				if (line == null)
				{
					System.err.println("Could not read line");
				}

				line = line.trim();
				items = line.split(separator);
				for (String item: items)
					w.chi.add(Double.parseDouble(item));
			}

			pswfc.wavefunctions.add(w);
		}

		// final closing tag
		line = reader.readLine();
		if (line == null)
		{
			System.err.println("Could not read line");
		}
		if (!line.equalsIgnoreCase("</PP_PSWFC>"))
		{
			System.err.println("Expecting '<PP_BETA>', '<PP_DIJ>' or '<PP_QIJ>'");
		}

		return pswfc;
	}

	private UPFData.PPRhoAtom ParseRhoAtom(BufferedReader reader) throws IOException
	{
		UPFData.PPRhoAtom rho = new UPFData.PPRhoAtom();
		rho.rho_at = new ArrayList<Double>();

		while (true)
		{
			String line = reader.readLine();
			if (line == null)
			{
				System.err.println("Could not read line inside rho atom");
			}

			line = line.trim();
			if (line.equalsIgnoreCase("</PP_RHOATOM>"))
			{
				break;
			}

			String[] items = line.split(separator);
			for (String item: items)
				rho.rho_at.add(Double.parseDouble(item));
		}

		return rho;
	}


	public UPFData ParseFile(BufferedReader reader) throws IOException
	{
		UPFData data = new UPFData();

		String line;
		while ((line = reader.readLine()) != null)
		{
			line = line.trim();		// remove leading and trailing white-space characters

			switch (line.toUpperCase())
			{
				//case "<PP_INFO>":
				//	//this.info = ParseInfo(reader);
				//	break;

				case "<PP_HEADER>":
					data.header = ParseHeader(reader);
					break;

				case "<PP_MESH>":
					data.mesh = ParseMesh(reader);
					break;

				case "<PP_NLCC>":
					data.nlcc = ParseNLCC(reader);
					break;

				case "<PP_LOCAL>":
					data.local = ParseLocal(reader);
					break;

				case "<PP_NONLOCAL>":
					data.nonLocal = ParseNonLocal(reader);
					break;

				case "<PP_PSWFC>":
					data.pswfc = ParsePSWFC(reader, data.header.natwfc, data.header.mesh);
					break;

				case "<PP_RHOATOM>":
					data.rhoAtom = ParseRhoAtom(reader);
					break;
			}
		}

		return data;
	}
}
